==> distortos-0.1.0.tar.xz: distortos-0.1.0/source/synchronization/MutexControlBlock.cpp <==
/**
 * \file
 * \brief MutexControlBlock class implementation
 *
 * \author Copyright (C) 2014-2015 Kamil Szczygiel http://www.distortec.com http://www.freddiechopin.info
 *
 * \par License
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "distortos/internal/synchronization/MutexControlBlock.hpp"

#include "distortos/internal/scheduler/getScheduler.hpp"
#include "distortos/internal/scheduler/Scheduler.hpp"

namespace distortos
{

namespace internal
{

namespace
{

/*---------------------------------------------------------------------------------------------------------------------+
| local types
+---------------------------------------------------------------------------------------------------------------------*/

/// PriorityInheritanceMutexControlBlockUnblockFunctor is a functor executed when unblocking a thread that is blocked on
/// a mutex with PriorityInheritance protocol
class PriorityInheritanceMutexControlBlockUnblockFunctor : public ThreadControlBlock::UnblockFunctor
{
public:

	/**
	 * \brief PriorityInheritanceMutexControlBlockUnblockFunctor's constructor
	 *
	 * \param [in] mutexControlBlock is a reference to MutexControlBlock that blocked the thread
	 */

	constexpr explicit PriorityInheritanceMutexControlBlockUnblockFunctor(const MutexControlBlock& mutexControlBlock) :
			mutexControlBlock_{mutexControlBlock}
	{

	}

	/**
	 * \brief PriorityInheritanceMutexControlBlockUnblockFunctor's function call operator
	 *
	 * If the wait for mutex was interrupted, requests update of boosted priority of current owner of the mutex. Pointer
	 * to MutexControlBlock with PriorityInheritance protocol which caused the thread to block is reset to nullptr.
	 *
	 * \param [in] threadControlBlock is a reference to ThreadControlBlock that is being unblocked
	 * \param [in] unblockReason is the reason of thread unblocking
	 */

	virtual void operator()(ThreadControlBlock& threadControlBlock,
			const ThreadControlBlock::UnblockReason unblockReason) const override
	{
		const auto owner = mutexControlBlock_.getOwner();

		// waiting for mutex was interrupted and some thread still holds it?
		if (unblockReason != ThreadControlBlock::UnblockReason::UnblockRequest && owner != nullptr)
			owner->updateBoostedPriority();

		threadControlBlock.setPriorityInheritanceMutexControlBlock(nullptr);
	}

private:

	/// reference to MutexControlBlock that blocked the thread
	const MutexControlBlock& mutexControlBlock_;
};

}	// namespace

/*---------------------------------------------------------------------------------------------------------------------+
| public functions
+---------------------------------------------------------------------------------------------------------------------*/

int MutexControlBlock::block()
{
	if (protocol_ == Protocol::PriorityInheritance)
		priorityInheritanceBeforeBlock();

	const PriorityInheritanceMutexControlBlockUnblockFunctor unblockFunctor {*this};
	return getScheduler().block(blockedList_, ThreadState::BlockedOnMutex,
			protocol_ == Protocol::PriorityInheritance ? &unblockFunctor : nullptr);
}

int MutexControlBlock::blockUntil(const TickClock::time_point timePoint)
{
	if (protocol_ == Protocol::PriorityInheritance)
		priorityInheritanceBeforeBlock();

	const PriorityInheritanceMutexControlBlockUnblockFunctor unblockFunctor {*this};
	return getScheduler().blockUntil(blockedList_, ThreadState::BlockedOnMutex, timePoint,
			protocol_ == Protocol::PriorityInheritance ? &unblockFunctor : nullptr);
}

uint8_t MutexControlBlock::getBoostedPriority() const
{
	if (protocol_ == Protocol::PriorityInheritance)
	{
		if (blockedList_.empty() == true)
			return 0;
		return blockedList_.front().getEffectivePriority();
	}

	if (protocol_ == Protocol::PriorityProtect)
		return priorityCeiling_;

	return 0;
}

void MutexControlBlock::lock()
{
	auto& scheduler = getScheduler();
	owner_ = &scheduler.getCurrentThreadControlBlock();

	if (protocol_ == Protocol::None)
		return;

	owner_->getOwnedProtocolMutexList().push_front(*this);

	if (protocol_ == Protocol::PriorityProtect)
		owner_->updateBoostedPriority();
}

void MutexControlBlock::unlockOrTransferLock()
{
	auto& oldOwner = *owner_;

	if (blockedList_.empty() == false)
		transferLock();
	else
		unlock();

	if (protocol_ == Protocol::None)
		return;

	oldOwner.updateBoostedPriority();

	if (owner_ == nullptr)
		return;

	owner_->updateBoostedPriority();
}

/*---------------------------------------------------------------------------------------------------------------------+
| private functions
+---------------------------------------------------------------------------------------------------------------------*/

void MutexControlBlock::priorityInheritanceBeforeBlock() const
{
	auto& currentThreadControlBlock = getScheduler().getCurrentThreadControlBlock();

	currentThreadControlBlock.setPriorityInheritanceMutexControlBlock(this);

	// calling thread is not yet on the blocked list, that's why it's effective priority is given explicitly
	owner_->updateBoostedPriority(currentThreadControlBlock.getEffectivePriority());
}

void MutexControlBlock::transferLock()
{
	owner_ = &blockedList_.front();	// pass ownership to the unblocked thread
	getScheduler().unblock(blockedList_.begin());

	if (node.isLinked() == false)
		return;

	MutexList::splice(owner_->getOwnedProtocolMutexList().begin(), MutexList::iterator{*this});

	if (protocol_ == Protocol::PriorityInheritance)
		owner_->setPriorityInheritanceMutexControlBlock(nullptr);
}

void MutexControlBlock::unlock()
{
	owner_ = nullptr;

	if (node.isLinked() == false)
		return;

	node.unlink();
}

}	// namespace internal

}	// namespace distortos

==> distortos-0.6.0.tar.xz: distortos-0.6.0/source/synchronization/MutexControlBlock.cpp <==
/**
 * \file
 * \brief MutexControlBlock class implementation
 *
 * \author Copyright (C) 2014-2017 Kamil Szczygiel http://www.distortec.com http://www.freddiechopin.info
 *
 * \par License
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "distortos/internal/synchronization/MutexControlBlock.hpp"

#include "distortos/internal/scheduler/getScheduler.hpp"
#include "distortos/internal/scheduler/Scheduler.hpp"

namespace distortos
{

namespace internal
{

namespace
{

/*---------------------------------------------------------------------------------------------------------------------+
| local types
+---------------------------------------------------------------------------------------------------------------------*/

/// PriorityInheritanceMutexControlBlockUnblockFunctor is a functor executed when unblocking a thread that is blocked on
/// a mutex with priorityInheritance protocol
class PriorityInheritanceMutexControlBlockUnblockFunctor : public UnblockFunctor
{
public:

	/**
	 * \brief PriorityInheritanceMutexControlBlockUnblockFunctor's constructor
	 *
	 * \param [in] mutexControlBlock is a reference to MutexControlBlock that blocked the thread
	 */

	constexpr explicit PriorityInheritanceMutexControlBlockUnblockFunctor(const MutexControlBlock& mutexControlBlock) :
			mutexControlBlock_{mutexControlBlock}
	{

	}

	/**
	 * \brief PriorityInheritanceMutexControlBlockUnblockFunctor's function call operator
	 *
	 * If the wait for mutex was interrupted, requests update of boosted priority of current owner of the mutex. Pointer
	 * to MutexControlBlock with priorityInheritance protocol which caused the thread to block is reset to nullptr.
	 *
	 * \param [in] threadControlBlock is a reference to ThreadControlBlock that is being unblocked
	 * \param [in] unblockReason is the reason of thread unblocking
	 */

	void operator()(ThreadControlBlock& threadControlBlock, const UnblockReason unblockReason) const override
	{
		const auto owner = mutexControlBlock_.getOwner();

		// waiting for mutex was interrupted and some thread still holds it?
		if (unblockReason != UnblockReason::unblockRequest && owner != nullptr)
			owner->updateBoostedPriority();

		threadControlBlock.setPriorityInheritanceMutexControlBlock(nullptr);
	}

private:

	/// reference to MutexControlBlock that blocked the thread
	const MutexControlBlock& mutexControlBlock_;
};

}	// namespace

/*---------------------------------------------------------------------------------------------------------------------+
| public functions
+---------------------------------------------------------------------------------------------------------------------*/

uint8_t MutexControlBlock::getBoostedPriority() const
{
	if (getProtocol() == Protocol::priorityInheritance)
	{
		if (blockedList_.empty() == true)
			return 0;
		return blockedList_.front().getEffectivePriority();
	}

	if (getProtocol() == Protocol::priorityProtect)
		return getPriorityCeiling();

	return 0;
}

/*---------------------------------------------------------------------------------------------------------------------+
| protected functions
+---------------------------------------------------------------------------------------------------------------------*/

int MutexControlBlock::doBlock()
{
	beforeBlock();

	const PriorityInheritanceMutexControlBlockUnblockFunctor unblockFunctor {*this};
	return getScheduler().block(blockedList_, ThreadState::blockedOnMutex,
			getProtocol() == Protocol::priorityInheritance ? &unblockFunctor : nullptr);
}

int MutexControlBlock::doBlockUntil(const TickClock::time_point timePoint)
{
	beforeBlock();

	const PriorityInheritanceMutexControlBlockUnblockFunctor unblockFunctor {*this};
	return getScheduler().blockUntil(blockedList_, ThreadState::blockedOnMutex, timePoint,
			getProtocol() == Protocol::priorityInheritance ? &unblockFunctor : nullptr);
}

void MutexControlBlock::doLock()
{
	auto& scheduler = getScheduler();
	owner_ = &scheduler.getCurrentThreadControlBlock();

	if (getProtocol() == Protocol::none)
		return;

	getOwner()->getOwnedProtocolMutexList().push_front(*this);

	if (getProtocol() == Protocol::priorityProtect)
		getOwner()->updateBoostedPriority();
}

void MutexControlBlock::doUnlockOrTransferLock()
{
	auto& oldOwner = *getOwner();

	if (blockedList_.empty() == false)
		doTransferLock();
	else
		doUnlock();

	if (getProtocol() == Protocol::none)
		return;

	oldOwner.updateBoostedPriority();

	if (getOwner() == nullptr)
		return;

	getOwner()->updateBoostedPriority();
}

/*---------------------------------------------------------------------------------------------------------------------+
| private functions
+---------------------------------------------------------------------------------------------------------------------*/

void MutexControlBlock::beforeBlock() const
{
	if (getProtocol() != Protocol::priorityInheritance)
		return;

	auto& currentThreadControlBlock = getScheduler().getCurrentThreadControlBlock();

	currentThreadControlBlock.setPriorityInheritanceMutexControlBlock(this);

	// calling thread is not yet on the blocked list, that's why it's effective priority is given explicitly
	getOwner()->updateBoostedPriority(currentThreadControlBlock.getEffectivePriority());
}

void MutexControlBlock::doTransferLock()
{
	owner_ = &blockedList_.front();	// pass ownership to the unblocked thread
	getScheduler().unblock(blockedList_.begin());

	if (node.isLinked() == false)
		return;

	MutexList::splice(getOwner()->getOwnedProtocolMutexList().begin(), MutexList::iterator{*this});

	if (getProtocol() == Protocol::priorityInheritance)
		getOwner()->setPriorityInheritanceMutexControlBlock(nullptr);
}

void MutexControlBlock::doUnlock()
{
	owner_ = nullptr;

	if (node.isLinked() == false)
		return;

	node.unlink();
}

}	// namespace internal

}	// namespace distortos
